#!/usr/bin/env python3
# coding=utf-8

"""
Reusable common utilities, interfaces and implementations for python projects related to collections.
"""

from vt.utils.commons.commons.collections.utils import get_first_true as get_first_true
from vt.utils.commons.commons.collections.utils import get_last_true as get_last_true
